from __future__ import annotations

import builtins
from unittest import mock

import pytest

from devenv import doctor
from tests.doctor.devenv.checks import passing_check


prompts_for_fix_yes: list[str] = [
    "y",
    "Y",
    "yes",
    "Yes",
    "yEs",
    "yeS",
    "YEs",
    "yES",
    "YES",
    "",
]

prompts_for_fix_no: list[str] = [
    "n",
    "N",
    "no",
    "No",
    "nO",
    "NO",
    "any other string",
]


@pytest.mark.parametrize("yes", prompts_for_fix_yes)
def test_prompt_for_fix_yes(yes: str) -> None:
    check = doctor.Check(passing_check)

    def fake_input(_: str) -> str:
        return yes

    with mock.patch.object(builtins, "input", fake_input):
        assert doctor.prompt_for_fix(check)


@pytest.mark.parametrize("no", prompts_for_fix_no)
def test_prompt_for_fix_no(no: str) -> None:
    check = doctor.Check(passing_check)

    def fake_input(_: str) -> str:
        return no

    with mock.patch.object(builtins, "input", fake_input):
        assert not doctor.prompt_for_fix(check)



from __future__ import annotations

import os
from unittest.mock import patch

import pytest

from devenv.lib.proc import run


def test_run_with_stdout() -> None:
    cmd = ("echo", "Hello, World!")
    expected_output = "Hello, World!"

    result = run(cmd, stdout=True)

    assert result == expected_output


def test_run_without_stdout() -> None:
    cmd = ("echo", "Hello, World!")

    result = run(cmd)

    assert result is None


def test_run_with_debug() -> None:
    cmd = ("echo", "Hello, World!")

    with patch("devenv.lib.proc.constants.DEBUG", True), patch(
        "devenv.lib.proc.xtrace"
    ) as mock_xtrace:
        run(cmd)

    mock_xtrace.assert_called_once_with(cmd)


def test_run_with_pathprepend(tmp_path: str) -> None:
    dummy_executable = os.path.join(tmp_path, "dummy_executable")
    with open(dummy_executable, "w") as f:
        f.write("#!/bin/sh\necho Hello, World!")
    os.chmod(dummy_executable, 0o755)

    cmd = ("dummy_executable",)
    pathprepend = tmp_path

    result = run(cmd, pathprepend=pathprepend, stdout=True)

    assert result == "Hello, World!"


def test_run_command_not_found() -> None:
    cmd = ("invalid_command",)

    with pytest.raises(RuntimeError):
        run(cmd, exit=False)


def test_run_command_not_found_with_exit() -> None:
    cmd = ("invalid_command",)

    with pytest.raises(SystemExit):
        run(cmd, exit=True)


def test_run_with_custom_env() -> None:
    cmd = ("sh", "-c", "printenv VAR1 && printenv VAR2")
    custom_env = {"VAR1": "value1", "VAR2": "value2"}

    result = run(cmd, env=custom_env, stdout=True)

    assert result == "value1\nvalue2"


def test_run_with_cwd(tmp_path: str) -> None:
    text = "Hello, World!"
    with open(os.path.join(tmp_path, "test.txt"), "w") as f:
        f.write(text)
    cmd = ("cat", "test.txt")
    cwd = tmp_path

    result = run(cmd, cwd=cwd, stdout=True)

    assert result == text


def test_run_command_failed() -> None:
    cmd = ("ls", "nonexistent_directory")

    with pytest.raises(RuntimeError):
        run(cmd, exit=False)


def test_run_command_failed_with_exit() -> None:
    cmd = ("ls", "nonexistent_directory")

    with pytest.raises(SystemExit):
        run(cmd, exit=True)



from __future__ import annotations

import os
import tempfile

from devenv.constants import CI
from devenv.constants import EXTERNAL_CONTRIBUTOR
from devenv.constants import home
from devenv.lib import fs
from devenv.lib import proc

# https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/githubs-ssh-key-fingerprints
fingerprints = """
github.com ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIOMqqnkVzrm0SdG6UOoqKLsabgH5C9okWi0dh2l9GKJl
github.com ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBEmKSENjQEezOmxkZMy7opKgwFB9nkt5YRrYMjNuG5N87uRgg6CLrbo5wAdT/y6v0mKV0U2w0WZ2YB/++Tpockg=
github.com ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQCj7ndNxQowgcQnjshcLrqPEiiphnt+VTTvDP6mHBL9j1aNUkY4Ue1gvwnGLVlOhGeYrnZaMgRK6+PKCUXaDbC7qtbW8gIkhL7aGCsOr/C56SJMy/BCZfxd1nWzAOxSDPgVsmerOBYfNqltV9/hWCqBywINIR+5dIg6JTJ72pcEpEjcYgXkE2YEFXV1JHnsKgbLWNlhScqb2UmyRkQyytRLtL+38TGxkxCflmO+5Z8CSSNY7GidjMIZ7Q4zMjA2n1nGrlTDkzwDCsw+wqFPGQA179cnfGWOWRVruj16z6XyvxvjJwbz0wQZ75XK5tKSb7FNyeIEs4TT4jk+S4dhPeAUC5y+bDYirYgM4GC7uEnztnZyaVWQ7B381AK4Qdrwt51ZqExKbQpTUNn+EjqoTwvqNj4kqx5QUCI0ThS/YkOxJCXmPUWZbhjpCg56i+2aB6CmK2JGhn57K5mj0MNdBXA4/WnwH6XoPWJzK5Nyu2zB3nAZp+S5hpQs+p1vN1/wsjk=
"""  # noqa


def check_ssh_authentication() -> bool:
    try:
        # The remote prints to stderr and exits with code 1 in all cases.
        proc.run(("sh", "-c", "ssh -T git@github.com 2>&1"), stdout=True)
    except RuntimeError as e:
        # https://docs.github.com/en/authentication/connecting-to-github-with-ssh/testing-your-ssh-connection
        if "You've successfully authenticated" not in str(e):
            print(e)
            return False
    return True


def check_sso_configuration() -> bool:
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            proc.run(
                (
                    "git",
                    "-C",
                    tmpdir,
                    "clone",
                    "--quiet",
                    "--depth=1",
                    "git@github.com:getsentry/private.git",
                )
            )
            return True
    except RuntimeError as e:
        # Failing to clone private repos under getsentry
        # means that SSO isn't configured for the ssh key.
        print(e)
    return False


def check_ssh_access() -> bool:
    if CI:
        return True
    ssh_auth = check_ssh_authentication()
    return (
        ssh_auth
        if EXTERNAL_CONTRIBUTOR
        else ssh_auth and check_sso_configuration()
    )


def add_to_known_hosts() -> None:
    os.makedirs(f"{home}/.ssh", exist_ok=True)
    fs.idempotent_add(f"{home}/.ssh/known_hosts", fingerprints)


def generate_and_configure_ssh_keypair() -> str:
    fs.idempotent_add(
        f"{home}/.ssh/config",
        """Host github.com
  User git
  Hostname github.com
  PreferredAuthentications publickey
  IdentityFile ~/.ssh/sentry-github""",
    )
    private_key_path = f"{home}/.ssh/sentry-github"
    if not os.path.exists(private_key_path):
        proc.run(
            (
                "ssh-keygen",
                "-t",
                "ed25519",
                "-a",
                "100",
                "-N",
                "",
                "-f",
                private_key_path,
            ),
            exit=True,
        )
    with open(f"{private_key_path}.pub") as f:
        return f.read().strip()



from __future__ import annotations

import contextlib
import os
import pathlib
from collections.abc import Generator


# TODO: replace with contextlib.chdir when we can use python3.11
@contextlib.contextmanager
def chdir(d: str | pathlib.Path) -> Generator[None, None, None]:
    curdir = os.getcwd()
    try:
        os.chdir(d)
        yield
    finally:
        os.chdir(curdir)



from __future__ import annotations

from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = set(["test", "broken"])
name = "broken check"


@checker
def check() -> tuple[bool, str]:
    # This is a broken check, it will raise an exception
    a = 1 / 0
    return True, f"{a}"


@fixer
def fix() -> tuple[bool, str]:
    return True, ""



from __future__ import annotations

import os

import pytest


def pytest_configure(config: pytest.Config) -> None:
    os.environ["CI"] = "1"
    os.environ["SHELL"] = "/bin/bash"
    os.environ["SENTRY_FORCE_PY"] = "1"



from __future__ import annotations

import os

from devenv.lib import fs
from devenv.lib import proc
from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = set()
name = "foo"


@checker
def check() -> tuple[bool, str]:
    if os.path.exists(f"{fs.gitroot()}/foo"):
        return True, ""
    return False, f"{fs.gitroot()}/foo doesn't exist"


@fixer
def fix() -> tuple[bool, str]:
    try:
        proc.run(
            (
                "/bin/bash",
                "-c",
                """
echo blah > foo
""",
            ),
            cwd=fs.gitroot(),
        )
        return True, ""
    except RuntimeError as e:
        return False, f"{e}"



from __future__ import annotations

from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = set(["test", "pass"])
name = "passing check"


@checker
def check() -> tuple[bool, str]:
    return True, ""


@fixer
def fix() -> tuple[bool, str]:
    return True, ""



from __future__ import annotations

import argparse
import importlib.util
import os
import sys
from collections.abc import Sequence

from devenv.constants import CI
from devenv.constants import DARWIN
from devenv.constants import EXTERNAL_CONTRIBUTOR
from devenv.constants import homebrew_bin
from devenv.lib import proc
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import ExitCode


def main(context: Context, argv: Sequence[str] | None = None) -> ExitCode:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "repo", type=str, help="the repository to fetch e.g., getsentry/sentry"
    )

    args = parser.parse_args(argv)
    code_root = context["code_root"]

    if args.repo in ["ops", "getsentry/ops"]:
        fetch(code_root, "getsentry/ops")
        fetch(code_root, "getsentry/terraform-modules", sync=False)
    elif args.repo in [
        "sentry",
        "getsentry",
        "getsentry/sentry",
        "getsentry/getsentry",
    ]:
        fetch(code_root, "getsentry/sentry", auth=CI is None, sync=False)

        print("Installing sentry's brew dependencies...")
        if CI:
            if DARWIN:
                # Installing everything from brew takes too much time,
                # and chromedriver cask flakes occasionally. Really all we need to
                # set up the devenv is colima and docker-cli.
                # This is also required for arm64 macOS GHA runners.
                # We manage colima, so just need to install docker + qemu here.
                proc.run(("brew", "install", "docker", "qemu"))
        else:
            proc.run(
                (f"{homebrew_bin}/brew", "bundle"), cwd=f"{code_root}/sentry"
            )

        proc.run(
            (sys.executable, "-P", "-m", "devenv", "sync"),
            cwd=f"{code_root}/sentry",
        )

        if not CI and not EXTERNAL_CONTRIBUTOR:
            fetch(code_root, "getsentry/getsentry")

        print(
            f"""

    All done! Please close this terminal window and start a fresh one.
    Sentry has been set up in {code_root}/sentry.
    cd into it and you should be able to run `sentry devserver`.

"""
        )
    else:
        fetch(code_root, args.repo)

    return 0


def fetch(
    coderoot: str, repo: str, auth: bool = True, sync: bool = True
) -> None:
    org, slug = repo.split("/")

    reporoot = f"{coderoot}/{slug}"

    if os.path.exists(reporoot):
        print(f"{reporoot} already exists")
        return

    print(f"fetching {repo} into {reporoot}")

    additional_args = (
        # git@ clones forces the use of cloning through SSH which is what we want,
        # though CI must clone open source repos via https (no git authentication)
        (f"git@github.com:{repo}",)
        if auth
        else (
            "--depth",
            "1",
            "--single-branch",
            f"--branch={os.environ['DEVENV_FETCH_BRANCH']}",
            f"https://github.com/{repo}",
        )
    )

    proc.run(("git", "-C", coderoot, "clone", *additional_args), exit=True)

    context_post_fetch = {
        "reporoot": reporoot,
        "repo": slug,
        "coderoot": coderoot,
    }

    # optional post-fetch, meant for recommended but not required defaults
    fp = f"{reporoot}/devenv/post_fetch.py"
    if os.path.exists(fp):
        spec = importlib.util.spec_from_file_location("post_fetch", fp)

        module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(module)  # type: ignore

        rc = module.main(context_post_fetch)
        if rc != 0:
            print(f"warning! failed running {fp} (code {rc})")

    if sync:
        proc.run((sys.executable, "-P", "-m", "devenv", "sync"), cwd=reporoot)


module_info = DevModuleInfo(
    action=main, name=__name__, command="fetch", help="Fetches a repository"
)



from __future__ import annotations

import os
import sys

from devenv.constants import CI
from devenv.constants import root
from devenv.lib import archive


# Forces download of the python interpreter even if the system interpreter is
# compatible with the requested version. By default, devenv uses system python
# on CI.
FORCE_PY = os.getenv("SENTRY_FORCE_PY", "").lower() in ("1", "true")


def _is_sys_compatible(version: str) -> bool:
    """Returns ``True`` if the current python interpreter is semver-compatible
    with the given version."""

    expected = tuple(int(i) for i in version.split("."))

    return (
        sys.version_info[:2] == expected[:2]
        and sys.version_info[2] >= expected[2]
    )


def get(python_version: str, url: str, sha256: str) -> str:
    unpack_into = f"{root}/pythons/{python_version}"

    if os.path.exists(f"{unpack_into}/python/bin/python3"):
        return f"{unpack_into}/python/bin/python3"

    if CI and not FORCE_PY and _is_sys_compatible(python_version):
        return sys.executable

    archive_file = archive.download(url, sha256)
    archive.unpack(archive_file, unpack_into)

    assert os.path.exists(f"{unpack_into}/python/bin/python3")
    return f"{unpack_into}/python/bin/python3"



from __future__ import annotations

import os
import platform
import shutil
from typing import Optional

from devenv.constants import homebrew_bin
from devenv.constants import root
from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc

_version = "1.1.1"
_sha256 = {
    f"volta-{_version}-macos-aarch64.tar.gz": "013d419550525fa6a212c2693798f9e2e65737e887e3604b08bc15a6be737e01",
    f"volta-{_version}-macos.tar.gz": "778ccaa943de8729d91e9df02a2767b470d97e8d5551faf6d89978db6f5f3c64",
    f"volta-{_version}-linux.tar.gz": "ab14e5d50ef836f8f43b56323cfbe7ba95a004bad05450b453c5b06a0b433d7b",
}


class UnexpectedPlatformError(Exception):
    pass


def determine_platform(unpack_into: str) -> str | None:
    system = platform.system()
    if system == "Linux":
        if platform.machine() == "x86_64":
            return f"volta-{_version}-linux.tar.gz"
        else:
            proc.run(("brew", "install", "volta"), exit=True)
            proc.run(
                ("sh", "-c", f"ln -sfn {homebrew_bin}/volta* {unpack_into}/"),
                exit=True,
            )
            return None
    elif system == "Darwin":
        suffix = "-aarch64" if platform.machine() == "arm64" else ""
        return f"volta-{_version}-macos{suffix}.tar.gz"
    else:
        raise UnexpectedPlatformError(f"Unexpected OS: {platform.platform()}")


def download_and_unpack_archive(name: str, unpack_into: str) -> None:
    url = (
        "https://github.com/volta-cli/volta/releases/download/"
        f"v{_version}/{name}"
    )

    archive_file = archive.download(url, _sha256[name])
    archive.unpack(archive_file, unpack_into)


def _install(unpack_into: str) -> None:
    name = determine_platform(unpack_into)
    if name is None:
        return
    download_and_unpack_archive(name, unpack_into)


def populate_volta_home_with_shims(unpack_into: str, volta_home: str) -> None:
    # executing volta -v will populate the VOLTA_HOME directory
    # with node/npm/yarn shims
    proc.run((f"{unpack_into}/volta-migrate",), env={"VOLTA_HOME": volta_home})
    version = proc.run(
        (f"{unpack_into}/volta", "-v"),
        env={"VOLTA_HOME": volta_home},
        stdout=True,
    )
    assert version == _version, (version, _version)


def uninstall(binroot: str, volta_home: str) -> None:
    shutil.rmtree(volta_home, ignore_errors=True)

    for executable in ("node", "npm", "npx", "yarn", "pnpm"):
        fp = f"{binroot}/{executable}"
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def install(reporoot: Optional[str] = "") -> None:
    if reporoot:
        binroot = fs.ensure_binroot(reporoot)
        VOLTA_HOME = f"{binroot}/volta-home"
    else:
        # compatibility with devenv <= 1.4.0
        binroot = f"{root}/bin"
        os.makedirs(binroot, exist_ok=True)
        VOLTA_HOME = f"{root}/volta"

    if (
        shutil.which("volta", path=binroot) == f"{binroot}/volta"
        and shutil.which("node", path=f"{VOLTA_HOME}/bin")
        == f"{VOLTA_HOME}/bin/node"
        and os.path.exists(f"{binroot}/node")
        # <= 1.6.0 had symlinks, now we have VOLTA_HOME shims
        and not os.path.islink(f"{binroot}/node")
    ):
        return

    print(f"installing volta {_version}...")
    uninstall(binroot, VOLTA_HOME)
    _install(binroot)
    populate_volta_home_with_shims(binroot, VOLTA_HOME)

    if not os.path.exists(f"{VOLTA_HOME}/bin/node"):
        raise SystemExit("Failed to install volta!")

    for executable in ("node", "npm", "npx", "yarn", "pnpm"):
        # This forces these repo-local bins to use the repo-local
        # volta home for full containment. Otherwise, the default
        # is to use ~/.volta.
        fs.write_script(
            f"{binroot}/{executable}",
            f"""#!/bin/sh
    exec /usr/bin/env VOLTA_HOME={VOLTA_HOME} {VOLTA_HOME}/bin/{executable} "$@"
""",
        )



from __future__ import annotations

import pathlib
import tarfile

import pytest

from devenv.lib import archive


@pytest.fixture
def tar(tmp_path: pathlib.Path) -> pathlib.Path:
    plain = tmp_path.joinpath("plain")
    plain.write_text("hello world\n")

    tar = tmp_path.joinpath("tar")

    with tarfile.open(tar, "w:tar") as tarf:
        tarf.add(plain, arcname="hello.txt")

    return tar


def test_unpack_tar(tar: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path.joinpath("dest")
    archive.unpack(str(tar), str(dest))
    assert dest.joinpath("hello.txt").read_text() == "hello world\n"



from __future__ import annotations

import os
import shutil
from enum import Enum
from typing import Optional

from devenv import pythons
from devenv.lib import config
from devenv.lib import fs
from devenv.lib import proc

VenvStatus = Enum(
    "VenvStatus", ("OK", "VERSION_MISMATCH", "NOT_PRESENT", "NOT_CONFIGURED")
)


# example venv configuration section:
#
# [venv.sentry-kube]
# python = 3.11.6
# requirements = k8s/cli/requirements.txt
# path = optional
# editable =
#   k8s/cli
#   k8s/cli/libsentrykube
# bins =
#   sentry-kube
#   sentry-kube-pop
#
# [venv.salt]
# python = 3.10.13
# requirements = salt/requirements.txt
# bins =
#   salt-ssh
#   ...
#
# example usage:
#
# venv_dir, python_version, requirements, editable_paths, bins = get(reporoot, "sentry-kube")
# url, sha256 = config.get_python(reporoot, python_version)
# ensure(path, python_version, url, sha256)
# sync(reporoot, venv_dir, requirements, editable_paths, bins)
def get(
    reporoot: str, name: str
) -> tuple[str, str, str, Optional[tuple[str, ...]], Optional[tuple[str, ...]]]:
    cfg = config.get_repo(reporoot)

    if not cfg.has_section(f"venv.{name}"):
        raise KeyError(f"section venv.{name} not found in repo config")

    venv = cfg[f"venv.{name}"]
    venv_dir = venv.get("path", f"{reporoot}/.venv-{name}")
    editable_paths = venv.get("editable", None)
    if editable_paths is not None:
        editable_paths = tuple(
            f"{reporoot}/{path}" for path in editable_paths.strip().split("\n")
        )

    bins = venv.get("bins", None)
    if bins is not None:
        bins = tuple(bins.strip().split("\n"))

    return (
        venv_dir,
        venv["python"],
        f"{reporoot}/{venv['requirements']}",
        editable_paths,
        bins,
    )


def sync(
    reporoot: str,
    venv_dir: str,
    requirements: str,
    editable_paths: Optional[tuple[str, ...]] = None,
    bins: Optional[tuple[str, ...]] = None,
) -> None:
    cmd: tuple[str, ...] = (
        f"{venv_dir}/bin/python",
        "-m",
        "pip",
        "--disable-pip-version-check",
        "--no-color",
        "--quiet",
        "--require-virtualenv",
        "install",
        "-r",
        requirements,
    )
    if editable_paths is not None:
        for path in editable_paths:
            cmd = (*cmd, "-e", path)
    proc.run(cmd)

    if bins is not None:
        binroot = fs.ensure_binroot(reporoot)
        for name in bins:
            fs.ensure_symlink(
                expected_src=f"{venv_dir}/bin/{name}", dest=f"{binroot}/{name}"
            )


def check(venv: str, python_version: str) -> VenvStatus:
    if not os.path.exists(f"{venv}/pyvenv.cfg"):
        return VenvStatus.NOT_PRESENT

    with open(f"{venv}/pyvenv.cfg", "r") as f:
        for line in f:
            if line.startswith("version"):
                venv_version = line.split("=")[1].strip()
                if venv_version != python_version:
                    return VenvStatus.VERSION_MISMATCH

    return VenvStatus.OK


def ensure(venv: str, python_version: str, url: str, sha256: str) -> None:
    venv_status = check(venv, python_version)
    if venv_status == VenvStatus.OK:
        return

    print(
        f"virtualenv doesn't exist or is using an outdated python, recreating at {venv}..."
    )
    if os.path.exists(venv):
        shutil.rmtree(venv)

    proc.run(
        (pythons.get(python_version, url, sha256), "-m", "venv", venv),
        exit=True,
    )

    with open(f"{venv}/.gitignore", "w") as f:
        f.write(
            """*
# automatically written by devenv
"""
        )



from __future__ import annotations

from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = set(["test", "bad"])
name = "bad fix"


@checker
def check() -> tuple[bool, str]:
    return True, ""


@fixer  # type: ignore  # intended error
def fix() -> tuple[str, str]:
    return "True", ""



from __future__ import annotations

import os
from shutil import which

from devenv.constants import DARWIN
from devenv.constants import homebrew_bin
from devenv.constants import homebrew_repo
from devenv.constants import INTEL_MAC
from devenv.constants import user
from devenv.lib import fs
from devenv.lib import proc


def create_dirs() -> None:
    dirs = homebrew_repo
    if INTEL_MAC:
        dirs = f"{dirs} /usr/local/Cellar /usr/local/Caskroom /usr/local/Frameworks /usr/local/bin /usr/local/etc /usr/local/include /usr/local/lib /usr/local/opt /usr/local/sbin /usr/local/share /usr/local/var"
    while True:
        try:
            proc.run(
                (
                    "sudo",
                    "bash",
                    "-c",
                    f"""
mkdir -p {dirs}
chown {user} {dirs}
""",
                ),
                exit=False,
            )
        except RuntimeError:
            continue
        break


def add_brew_to_shellrc() -> None:
    shellrc = fs.shellrc()
    fs.idempotent_add(
        shellrc,
        f"""
eval "$({homebrew_bin}/brew shellenv)"
""",
    )


def install() -> None:
    if not DARWIN:
        return

    # idempotency: skip if brew is on the executing shell's path
    if which("brew") is not None:
        return

    print("You may be prompted for your password to install homebrew.")
    create_dirs()
    proc.run(
        (
            "git",
            "-C",
            homebrew_repo,
            "clone",
            # homebrew works without any previous history as updating is just pulling
            "--depth=1",
            "https://github.com/Homebrew/brew",
            ".",
        )
    )
    if INTEL_MAC:
        os.symlink(f"{homebrew_repo}/bin/brew", f"{homebrew_bin}/brew")
    add_brew_to_shellrc()



from __future__ import annotations

import importlib.util
import os
from collections.abc import Sequence

from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import require_repo


@require_repo
def main(context: Context, argv: Sequence[str] | None = None) -> int:
    repo = context["repo"]
    assert repo is not None

    if not os.path.exists(f"{repo.config_path}/sync.py"):
        print(f"{repo.config_path}/sync.py not found!")
        return 1

    spec = importlib.util.spec_from_file_location(
        "sync", f"{repo.config_path}/sync.py"
    )

    module = importlib.util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore

    context_compat = {
        "reporoot": repo.path,
        "repo": repo.name,
        "coderoot": context.get("code_root"),
    }
    return module.main(context_compat)  # type: ignore


module_info = DevModuleInfo(
    action=main, name=__name__, command="sync", help="Resyncs the environment."
)



from __future__ import annotations

import os.path
from configparser import ConfigParser

from devenv.lib.config import get_config
from devenv.lib.fs import ensure_binroot


class Repository:
    def __init__(self, root: str) -> None:
        self.path = root
        self.name = os.path.basename(root)

        # .devenv may exist but not be config_path because .devenv/bin is hard-coded, so we check for config.ini
        if os.path.exists(f"{root}/.devenv/config.ini"):
            self.config_path = f"{root}/.devenv"
        # devenv _may_ exist but not be config_path because devenv has a devenv module, so we check for config.ini
        elif os.path.exists(f"{root}/devenv/config.ini"):
            self.config_path = f"{root}/devenv"
        else:
            # new default config_path is .devenv
            self.config_path = f"{root}/.devenv"

    def __repr__(self) -> str:
        return f"Repository: {self.name}"

    def ensure_binroot(self) -> None:
        ensure_binroot(self.path)

    def config(self) -> ConfigParser:
        return get_config(f"{self.config_path}/config.ini")



from __future__ import annotations

import os

import pytest

from devenv import doctor
from devenv.lib.repository import Repository


def test_load_checks_no_checks() -> None:
    assert doctor.load_checks(Repository("not a real repository"), set()) == []


def test_load_checks_test_checks(capsys: pytest.CaptureFixture[str]) -> None:
    loaded_checks = doctor.load_checks(
        Repository(os.path.join(os.path.dirname(__file__))), set()
    )
    loaded_check_names = [check.name for check in loaded_checks]
    assert len(loaded_check_names) == 5
    assert "passing check" in loaded_check_names
    assert "failing check" in loaded_check_names
    assert "failing check with msg" in loaded_check_names
    assert "broken check" in loaded_check_names
    assert "broken fix" in loaded_check_names
    captured = capsys.readouterr()

    assert (
        captured.out
        == """⚠️ Skipping bad_check: `check(...)` should return a tuple of (bool, str)
⚠️ Skipping bad_fix: `fix(...)` should return a tuple of (bool, str)
⚠️ Skipping no_check: must have a `check` function
⚠️ Skipping no_name: missing the `name` attribute
⚠️ Skipping no_tags: missing the `tags` attribute\n"""
    )


def test_load_checks_only_passing_tag() -> None:
    loaded_checks = doctor.load_checks(
        Repository(os.path.join(os.path.dirname(__file__))), {"pass"}
    )
    loaded_check_names = [check.name for check in loaded_checks]
    assert len(loaded_check_names) == 1
    assert "passing check" in loaded_check_names


def test_load_checks_only_failing_tag() -> None:
    loaded_checks = doctor.load_checks(
        Repository(os.path.join(os.path.dirname(__file__))), {"fail"}
    )
    loaded_check_names = [check.name for check in loaded_checks]
    assert len(loaded_check_names) == 2
    assert "failing check" in loaded_check_names
    assert "failing check with msg" in loaded_check_names


def test_load_checks_passing_and_failing_tag() -> None:
    loaded_checks = doctor.load_checks(
        Repository(os.path.join(os.path.dirname(__file__))), {"pass", "fail"}
    )
    loaded_check_names = [check.name for check in loaded_checks]
    assert len(loaded_check_names) == 0


def test_load_checks_test_tag() -> None:
    loaded_checks = doctor.load_checks(
        Repository(os.path.join(os.path.dirname(__file__))), {"test"}
    )
    loaded_check_names = [check.name for check in loaded_checks]
    assert len(loaded_check_names) == 5
    assert "passing check" in loaded_check_names
    assert "failing check" in loaded_check_names
    assert "failing check with msg" in loaded_check_names
    assert "broken check" in loaded_check_names
    assert "broken fix" in loaded_check_names



from __future__ import annotations

import argparse
import importlib.util
import typing
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from pkgutil import walk_packages
from types import ModuleType
from typing import Dict
from typing import List

from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import require_repo
from devenv.lib.repository import Repository
from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer


class Check:
    """
    A check is a module with the following attributes:
    - name: str
    - tags: set[str]
    - check: Callable[[], tuple[bool, str]]
    - fix: Callable[[], tuple[bool, str]]

    The check function should return a tuple of (ok, msg).
    The fix function should return a tuple of (ok, msg).

    The check and fix functions are wrapped with the checker and fixer decorators.
    """

    name: str
    tags: set[str]
    check: Callable[[], tuple[bool, str]]
    fix: Callable[[], tuple[bool, str]]

    def __init__(self, module: ModuleType):
        # Check that the module has the required attributes.
        assert hasattr(module, "name"), "missing the `name` attribute"
        assert isinstance(
            module.name, str
        ), "the `name` attribute should be a str"
        self.name = module.name

        assert hasattr(module, "tags"), "missing the `tags` attribute"
        assert isinstance(
            module.tags, set
        ), "the `tags` attribute should be a set"
        self.tags = module.tags

        # Check that the module has the check and fix functions.
        assert hasattr(module, "check"), "must have a `check` function"
        assert callable(
            module.check
        ), "the `check` attribute must be a function"
        check_hints = typing.get_type_hints(module.check)
        assert (
            check_hints["return"] == tuple[bool, str]
        ), "`check(...)` should return a tuple of (bool, str)"
        self.check = checker(module.check)

        assert hasattr(module, "fix"), "must have a `fix` function"
        assert callable(module.fix), "the `fix` attribute should be a function"
        fix_hints = typing.get_type_hints(module.fix)
        assert (
            fix_hints["return"] == tuple[bool, str]
        ), "`fix(...)` should return a tuple of (bool, str)"
        self.fix = fixer(module.fix)

        super().__init__()


def load_checks(repo: Repository, match_tags: set[str]) -> List[Check]:
    """
    Load all checks from the checks directory.
    Optionally filter by tags.
    If a check doesn't have the required attributes, skip it.
    """
    checks: list[Check] = []

    for module_finder, module_name, _ in walk_packages(
        (f"{repo.config_path}/checks",)
    ):
        module_spec = module_finder.find_spec(module_name, None)

        # it "should be" impossible to fail these:
        assert module_spec is not None, module_name
        assert module_spec.loader is not None, module_name

        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        try:
            check = Check(module)
        except AssertionError as e:
            print(f"⚠️ Skipping {module_name}: {e}")
            continue
        if match_tags and not check.tags.issuperset(match_tags):
            continue
        checks.append(check)
    return checks


def run_checks(
    checks: List[Check],
    executor: ThreadPoolExecutor,
    skip: Iterable[Check] = (),
) -> Dict[Check, tuple[bool, str]]:
    """
    Run checks in parallel, and return a dict of results.
    Results are a tuple of (ok, msg).
    """
    futures: dict[Check, Future[tuple[bool, str]]] = {}
    results: dict[Check, tuple[bool, str]] = {}
    for check in checks:
        if check in skip:
            print(f"\t⏭️  Skipped {check.name}".expandtabs(4))
            continue
        futures[check] = executor.submit(check.check)
    for check, future in futures.items():
        try:
            results[check] = future.result()
        except Exception as e:
            results[check] = (False, f"Check threw a runtime exception: {e}")
    return results


def filter_failing_checks(
    results: Dict[Check, tuple[bool, str]]
) -> List[Check]:
    """Print a report of the results, and return a list of failing checks."""
    failing_checks: list[Check] = []
    for check, result in results.items():
        ok, msg = result
        if ok:
            print(f"\t✅ check: {check.name}".expandtabs(4))
            continue
        print(f"\t❌ check: {check.name}{msg}".expandtabs(4))
        failing_checks.append(check)
    return failing_checks


def prompt_for_fix(check: Check) -> bool:
    """Prompt the user to attempt a fix."""
    return input(
        f"\t\tDo you want to attempt to fix {check.name}? (Y/n): ".expandtabs(4)
    ).lower() in {"y", "yes", ""}


def attempt_fix(check: Check, executor: ThreadPoolExecutor) -> tuple[bool, str]:
    """Attempt to fix a check, and return a tuple of (ok, msg)."""
    future = executor.submit(check.fix)
    try:
        return future.result()
    except Exception as e:
        return False, f"Fix threw a runtime exception: {e}"


@require_repo
def main(context: Context, argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tag",
        type=str,
        action="append",
        help="Used to match a subset of checks.",
    )
    parser.add_argument(
        "--check-only", action="store_true", help="Do not run fixers."
    )
    args = parser.parse_args(argv)

    repo = context["repo"]
    assert repo is not None

    match_tags: set[str] = set(args.tag if args.tag else ())
    checks = load_checks(repo, match_tags)

    if not checks:
        print(f"No checks found for tags: {args.tag}")
        return 1

    # We run every check on a separate thread, aggregate the results,
    # attempt any fixes, then recheck and provide a final report.
    executor = ThreadPoolExecutor(max_workers=len(checks))
    print(f"Running checks: {', '.join(f'{c.name}' for c in checks)}")

    results = run_checks(checks, executor)

    failing_checks = filter_failing_checks(results)

    if not failing_checks:
        print("\nLooks good to me.")
        executor.shutdown()
        return 0
    else:
        if args.check_only:
            return 1

    print("\nThe following problems have been identified:")
    skip: list[Check] = []
    for check in failing_checks:
        print(f"\t❌ {check.name}".expandtabs(4))
        # Prompt for fixes one by one, so the user can decide to skip a fix.
        if prompt_for_fix(check):
            ok, msg = attempt_fix(check, executor)
            if ok:
                print(f"\t\t✅ fix: {check.name}".expandtabs(4))
            else:
                print(f"\t\t❌ fix: {check.name}{msg}".expandtabs(4))
        else:
            print(f"\t\t⏭️  Skipping {check.name}".expandtabs(4))
            skip.append(check)

    print("\nChecking that fixes worked as expected...")
    results = run_checks(failing_checks, executor, skip=skip)

    executor.shutdown()

    all_ok = len(filter_failing_checks(results)) == 0

    if all_ok:
        print("\nLooks good to me now!")
        return 0

    print("\nSorry I couldn't fix it... ask for help in #discuss-dev-infra.")
    return 1


module_info = DevModuleInfo(
    action=main,
    name=__name__,
    command="doctor",
    help="Diagnose common issues, and optionally try to fix them.",
)



from __future__ import annotations

from pathlib import Path
from subprocess import CalledProcessError
from subprocess import PIPE
from subprocess import run as subprocess_run
from typing import Literal
from typing import overload

from devenv import constants
from devenv.constants import home
from devenv.constants import homebrew_bin
from devenv.constants import root
from devenv.constants import shell_path
from devenv.constants import user_environ

base_path = f"{root}/bin:{homebrew_bin}:{user_environ['PATH']}"
base_env = {"PATH": base_path, "HOME": home, "SHELL": shell_path}


def quote(cmd: tuple[str, ...]) -> str:
    """convert a command to bash-compatible form"""
    from pipes import quote

    return " ".join(quote(arg) for arg in cmd)


def xtrace(cmd: tuple[str, ...]) -> None:
    """Print a commandline, similar to how xtrace does."""

    teal = "\033[36m"
    reset = "\033[m"
    bold = "\033[1m"

    print(f"+ {teal}${reset} {bold}{quote(cmd)}{reset}")


@overload
def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    stdout: Literal[False] = False,
) -> None:
    ...


@overload
def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    stdout: Literal[True],
) -> str:
    ...


def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    stdout: bool = False,
) -> str | None:
    _stdout = PIPE if stdout else None
    del stdout

    if env is None:
        env = {}
    env = {**constants.user_environ, **base_env, **env}

    if pathprepend:
        env["PATH"] = f"{pathprepend}:{env['PATH']}"

    if constants.DEBUG:
        xtrace(cmd)
    try:
        proc = subprocess_run(cmd, check=True, stdout=_stdout, cwd=cwd, env=env)
    except FileNotFoundError as e:
        # This is reachable if the command isn't found.
        if exit:
            raise SystemExit(f"{e}") from None
        else:
            raise RuntimeError(f"{e}") from None
    except CalledProcessError as e:
        detail = f"Command `{quote(e.cmd)}` failed! (code {e.returncode})"
        if _stdout:
            detail += f"""
stdout:
{"" if e.stdout is None else e.stdout.decode()}
"""
        if exit:
            raise SystemExit(detail) from None
        else:
            raise RuntimeError(detail) from None
    else:
        if _stdout:
            return proc.stdout.decode().strip()
        return None



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

import pytest

from devenv import doctor
from tests.doctor.devenv.checks import broken_check
from tests.doctor.devenv.checks import failing_check
from tests.doctor.devenv.checks import failing_check_with_msg
from tests.doctor.devenv.checks import passing_check


def test_run_checks_no_checks() -> None:
    assert doctor.run_checks([], ThreadPoolExecutor()) == {}


def test_run_checks_one_passing_check() -> None:
    check = doctor.Check(passing_check)
    assert doctor.run_checks([check], ThreadPoolExecutor()) == {
        check: (True, "")
    }


def test_run_checks_one_failing_check() -> None:
    check = doctor.Check(failing_check)
    assert doctor.run_checks([check], ThreadPoolExecutor()) == {
        check: (False, "")
    }


def test_run_checks_one_failing_check_with_msg() -> None:
    check = doctor.Check(failing_check_with_msg)
    assert doctor.run_checks([check], ThreadPoolExecutor()) == {
        check: (False, "check failed")
    }


def test_run_checks_one_passing_and_one_failing_check() -> None:
    first_check = doctor.Check(passing_check)
    second_check = doctor.Check(failing_check)
    assert doctor.run_checks(
        [first_check, second_check], ThreadPoolExecutor()
    ) == {first_check: (True, ""), second_check: (False, "")}


def test_run_checks_skip(capsys: pytest.CaptureFixture[str]) -> None:
    first_check = doctor.Check(passing_check)
    second_check = doctor.Check(failing_check)
    assert doctor.run_checks(
        [first_check, second_check], ThreadPoolExecutor(), skip=[second_check]
    ) == {first_check: (True, "")}
    captured = capsys.readouterr()
    assert captured.out == "    ⏭️  Skipped failing check\n"


def test_run_checks_multiple_failing_checks() -> None:
    first_check = doctor.Check(failing_check)
    second_check = doctor.Check(failing_check_with_msg)
    assert doctor.run_checks(
        [first_check, second_check], ThreadPoolExecutor()
    ) == {first_check: (False, ""), second_check: (False, "check failed")}


def test_run_checks_broken_check() -> None:
    check = doctor.Check(broken_check)
    assert doctor.run_checks([check], ThreadPoolExecutor()) == {
        check: (False, "Check threw a runtime exception: division by zero")
    }



from __future__ import annotations

from collections.abc import Callable


def checker(
    f: Callable[[], tuple[bool, str]]
) -> Callable[[], tuple[bool, str]]:
    return f


def fixer(f: Callable[[], tuple[bool, str]]) -> Callable[[], tuple[bool, str]]:
    return f



from __future__ import annotations

from devenv.constants import SYSTEM_MACHINE
from devenv.lib import config
from devenv.lib import tenv
from devenv.lib import venv


def main(context: dict[str, str]) -> int:
    reporoot = context["reporoot"]

    cfg = config.get_repo(reporoot)

    tenv.install(
        cfg["tenv"]["version"],
        cfg["tenv"][SYSTEM_MACHINE],
        cfg["tenv"][f"{SYSTEM_MACHINE}_sha256"],
        reporoot,
    )

    name = "foo"
    venv_dir, python_version, requirements, editable_paths, bins = venv.get(
        reporoot, name
    )
    url, sha256 = config.get_python(reporoot, python_version)
    print(f"ensuring {name} venv at {venv_dir}...")
    venv.ensure(venv_dir, python_version, url, sha256)

    print(f"syncing {name} with {requirements}...")
    venv.sync(reporoot, venv_dir, requirements, editable_paths, bins)

    return 0



from __future__ import annotations

import os
import shutil
import sys

from devenv.constants import MACHINE
from devenv.constants import root
from devenv.constants import shell
from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc

_version = "2.32.3"

_sha256 = {
    "direnv.darwin-amd64": "6ff42606edb38ffce5e1a3f4a1c69401e42a7c49b8bdc4ddafd705bc770bd15c",
    "direnv.darwin-arm64": "dd053025ecae958118b3db2292721464e68da4fb319b80905a4cebba5ba9f069",
    "direnv.linux-amd64": "c0443e4600e4b93ec253e663fff18b5fb490cc555b1fc72f0ab22beaa2bb9810",
    "direnv.linux-arm64": "785684d2b228bcb1e3977545f0215966a5124cdffcf7a01a90cf1bf1aee71483",
}


def install() -> None:
    direnv_path = f"{root}/bin/direnv"

    if shutil.which("direnv") == direnv_path:
        return

    machine = "arm64" if MACHINE == "arm64" else "amd64"
    name = f"direnv.{sys.platform}-{machine}"
    url = (
        "https://github.com/direnv/direnv/releases/download"
        f"/v{_version}/{name}"
    )

    archive.download(url, _sha256[name], dest=direnv_path)
    os.chmod(direnv_path, 0o775)

    version = proc.run((direnv_path, "version"), stdout=True)
    assert version == _version, (version, _version)

    fs.idempotent_add(
        fs.shellrc(),
        f"""
eval "$(direnv hook {shell})"
""",
    )



from __future__ import annotations

import os
import shutil
import tempfile

from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc


def _install(url: str, sha256: str, into: str) -> None:
    TENV_ROOT = f"{into}/tenv-root"
    os.makedirs(f"{TENV_ROOT}/bin", exist_ok=True)

    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")
        archive.unpack(archive_file, tmpd)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/terraform", f"{TENV_ROOT}/bin/terraform")
        os.replace(f"{tmpd}/tf", f"{TENV_ROOT}/bin/tf")
        os.replace(f"{tmpd}/terragrunt", f"{TENV_ROOT}/bin/terragrunt")
        os.replace(f"{tmpd}/tenv", f"{TENV_ROOT}/bin/tenv")

        # those all need to go inside a bin instead of like, TENV_ROOT/terraform
        # because tenv wants to mkdir that

    # These shims make sure we're executing with our custom TENV_ROOT,
    # otherwise there's potential for collision with ~/.tenv.
    # This isolation also makes uninstallation safe.
    fs.write_script(
        f"{into}/tenv",
        f"""#!/bin/sh
exec /usr/bin/env TENV_ROOT={TENV_ROOT} {TENV_ROOT}/bin/tenv "$@"
""",
    )
    fs.write_script(
        f"{into}/terraform",
        f"""#!/bin/sh
exec /usr/bin/env TENV_ROOT={TENV_ROOT} {TENV_ROOT}/bin/terraform "$@"
""",
    )
    fs.write_script(
        f"{into}/terragrunt",
        f"""#!/bin/sh
exec /usr/bin/env TENV_ROOT={TENV_ROOT} {TENV_ROOT}/bin/terragrunt "$@"
""",
    )


def uninstall(binroot: str) -> None:
    for d in (f"{binroot}/tenv-root",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (
        f"{binroot}/tenv",
        f"{binroot}/terraform",
        f"{binroot}/terragrunt",
    ):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def _version(binpath: str) -> str:
    stdout = proc.run((binpath, "version"), stdout=True)
    # tenv version v1.3.0
    return stdout.split()[-1]


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    binroot = fs.ensure_binroot(reporoot)
    binpath = f"{binroot}/tenv"

    if shutil.which("tenv", path=binroot) == binpath:
        installed_version = _version(binpath)
        if version == installed_version:
            return
        print(f"installed tenv {installed_version} is outdated!")

    print(f"installing tenv {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    installed_version = _version(binpath)
    if version != installed_version:
        raise SystemExit("Failed to install tenv {version}!")



from __future__ import annotations

import os
import shutil
import tempfile

from devenv.constants import root
from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")
        archive.unpack(archive_file, tmpd)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/google-cloud-sdk", f"{into}/google-cloud-sdk")

    # I think gcloud will support 3.11 for quite some time, but this could
    # change to be more flexible in the future.
    # We run gcloud on 3.11 in gocd prod and it's been great,
    # and we may as well reuse devenv's internal python.
    fs.write_script(
        f"{into}/gcloud",
        f"""#!/bin/sh
exec /usr/bin/env CLOUDSDK_PYTHON={root}/python/bin/python3 PATH={into}/google-cloud-sdk/bin:$PATH gcloud "$@"
""",
    )
    fs.write_script(
        f"{into}/gsutil",
        f"""#!/bin/sh
exec /usr/bin/env CLOUDSDK_PYTHON={root}/python/bin/python3 PATH={into}/google-cloud-sdk/bin:$PATH gsutil "$@"
""",
    )


def uninstall(binroot: str) -> None:
    for d in (f"{binroot}/google-cloud-sdk",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (f"{binroot}/gcloud", f"{binroot}/gsutil"):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    binroot = fs.ensure_binroot(reporoot)

    if (
        shutil.which("gcloud", path=binroot) == f"{binroot}/gcloud"
        and shutil.which("gsutil", path=binroot) == f"{binroot}/gsutil"
        and shutil.which("gke-gcloud-auth-plugin", path=binroot)
        == f"{binroot}/gke-gcloud-auth-plugin"
    ):
        with open(f"{binroot}/google-cloud-sdk/VERSION", "r") as f:
            installed_version = f.read().strip()
            if version == installed_version:
                return
            print(f"installed gcloud {installed_version} is outdated!")

    print(f"installing gcloud {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    proc.run(
        (
            f"{binroot}/gcloud",
            "components",
            "install",
            "-q",
            "--verbosity=error",
            "gke-gcloud-auth-plugin",
        )
    )

    fs.ensure_symlink(
        f"{binroot}/google-cloud-sdk/bin/gke-gcloud-auth-plugin",
        f"{binroot}/gke-gcloud-auth-plugin",
    )

    stdout = proc.run((f"{binroot}/gcloud", "--version"), stdout=True)
    if "gke-gcloud-auth-plugin" not in stdout:
        raise SystemExit("Failed to install gcloud {version}!")



from __future__ import annotations

import os
import pathlib
import subprocess

import pytest

from devenv.lib.fs import gitroot
from tests.utils import chdir


def test_gitroot(tmp_path: pathlib.Path) -> None:
    subprocess.run(("git", "init", f"{tmp_path}"))

    with chdir(tmp_path):
        assert os.path.samefile(tmp_path, gitroot())
        assert os.path.isdir(f"{gitroot()}/.git")


def test_gitroot_cd(tmp_path: pathlib.Path) -> None:
    subprocess.run(("git", "init", f"{tmp_path}"))
    os.mkdir(f"{tmp_path}/foo")

    _gitroot = gitroot(cd=f"{tmp_path}/foo")
    assert os.path.samefile(f"{tmp_path}", _gitroot)
    assert os.path.isdir(f"{_gitroot}/.git")


def test_no_gitroot(tmp_path: pathlib.Path) -> None:
    with pytest.raises(RuntimeError):
        with chdir(tmp_path):
            gitroot()



from __future__ import annotations

from typing import TypedDict

from devenv.lib.repository import Repository


class Context(TypedDict):
    config_path: str
    code_root: str
    repo: Repository | None



from __future__ import annotations

import os
from unittest.mock import patch

from devenv import main
from devenv.lib.config import read_config


def test(tmp_path: str) -> None:
    with patch("devenv.constants.CI", True):
        config_path = f"{tmp_path}/.config/sentry-devenv/config.ini"
        coderoot = f"{tmp_path}/code"

        assert not os.path.exists(config_path)
        main.devenv(
            ("/path/to/argv0", "bootstrap", "-d", f"coderoot:{coderoot}"),
            config_path,
        )
        assert os.path.exists(config_path)

        config = read_config(config_path)
        assert config.get("devenv", "coderoot") == coderoot



from __future__ import annotations

import os
from unittest.mock import patch

import pytest

from devenv.lib.github import add_to_known_hosts
from devenv.lib.github import check_ssh_access
from devenv.lib.github import check_ssh_authentication
from devenv.lib.github import fingerprints
from devenv.lib.github import generate_and_configure_ssh_keypair


def test_generate_and_configure_ssh_keypair(tmp_path: str) -> None:
    home = tmp_path
    private_key_path = f"{home}/.ssh/sentry-github"
    public_key_path = f"{private_key_path}.pub"

    with patch("devenv.lib.github.home", home), patch(
        "devenv.lib.github.fs.idempotent_add"
    ) as mock_idempotent_add, patch(
        "devenv.lib.github.os.path.exists", return_value=False
    ) as mock_exists, patch(
        "devenv.lib.github.proc.run"
    ), patch(
        "builtins.open"
    ) as mock_open:
        mock_open.return_value.__enter__.return_value.read.return_value = (
            "public_key"
        )

        result = generate_and_configure_ssh_keypair()

        assert result == "public_key"
        mock_idempotent_add.assert_called_once_with(
            f"{home}/.ssh/config",
            """Host github.com
  User git
  Hostname github.com
  PreferredAuthentications publickey
  IdentityFile ~/.ssh/sentry-github""",
        )
        mock_exists.assert_called_once_with(private_key_path)
        mock_open.assert_called_once_with(public_key_path)


def test_check_ssh_authentication_success() -> None:
    with patch(
        "devenv.lib.github.proc.run",
        side_effect=[RuntimeError("You've successfully authenticated")],
    ):
        assert check_ssh_authentication()


def test_check_ssh_authentication_failure() -> None:
    with patch(
        "devenv.lib.github.proc.run",
        side_effect=[RuntimeError("Authentication failed")],
    ):
        assert not check_ssh_authentication()


def test_check_ssh_access_success() -> None:
    with patch("devenv.lib.github.CI", False), patch(
        "devenv.lib.github.check_ssh_authentication", return_value=True
    ), patch("devenv.lib.github.check_sso_configuration", return_value=True):
        assert check_ssh_access()


def test_check_ssh_access_failure_no_ssh() -> None:
    with patch("devenv.lib.github.CI", False), patch(
        "devenv.lib.github.check_ssh_authentication", return_value=False
    ), patch("devenv.lib.github.check_sso_configuration", return_value=True):
        assert not check_ssh_access()


def test_check_ssh_access_failure_no_sso() -> None:
    with patch("devenv.lib.github.CI", False), patch(
        "devenv.lib.github.check_ssh_authentication", return_value=True
    ), patch("devenv.lib.github.check_sso_configuration", return_value=False):
        assert not check_ssh_access()


@pytest.mark.parametrize("ssh", [True, False])
@pytest.mark.parametrize("sso", [True, False])
def test_check_ssh_access_success_ci(ssh: bool, sso: bool) -> None:
    with patch("devenv.lib.github.CI", True), patch(
        "devenv.lib.github.check_ssh_authentication", return_value=ssh
    ), patch("devenv.lib.github.check_sso_configuration", return_value=sso):
        assert check_ssh_access()


@pytest.mark.parametrize("sso", [True, False])
def test_check_ssh_access_success_external_contributor(sso: bool) -> None:
    with patch("devenv.lib.github.CI", False), patch(
        "devenv.lib.github.EXTERNAL_CONTRIBUTOR", "1"
    ), patch(
        "devenv.lib.github.check_ssh_authentication", return_value=True
    ), patch(
        "devenv.lib.github.check_sso_configuration", return_value=sso
    ) as mock_check_sso_configuration:
        assert check_ssh_access()

    # If the user is an external contributor, we should not check for SSO
    mock_check_sso_configuration.assert_not_called()


def test_check_ssh_access_failure_external_contributor() -> None:
    with patch("devenv.lib.github.CI", False), patch(
        "devenv.lib.github.EXTERNAL_CONTRIBUTOR", "1"
    ), patch("devenv.lib.github.check_ssh_authentication", return_value=False):
        assert not check_ssh_access()


def test_add_to_known_hosts(tmp_path: str) -> None:
    with patch("devenv.lib.github.home", tmp_path), patch.object(
        os, "makedirs"
    ) as mock_makedirs, patch(
        "devenv.lib.brew.fs.idempotent_add"
    ) as mock_idempotent_add:
        add_to_known_hosts()

    mock_makedirs.assert_called_once_with(f"{tmp_path}/.ssh", exist_ok=True)
    mock_idempotent_add.assert_called_once_with(
        f"{tmp_path}/.ssh/known_hosts", fingerprints
    )



from __future__ import annotations

import importlib.metadata
import os
import platform
import pwd
import typing

version = importlib.metadata.version("sentry_devenv")

CI = os.getenv("CI")
SYSTEM = platform.system().lower()
MACHINE = platform.machine()
DARWIN = SYSTEM == "darwin"
INTEL_MAC = DARWIN and (MACHINE == "x86_64")
SHELL_UNSET = "(SHELL unset)"
DEBUG = os.getenv("SNTY_DEVENV_DEBUG", os.getenv("DEBUG", ""))
EXTERNAL_CONTRIBUTOR = os.getenv("SENTRY_EXTERNAL_CONTRIBUTOR", "")

# for matching to download urls in repo config
SYSTEM_MACHINE = f"{SYSTEM}_{MACHINE}"

struct_passwd = pwd.getpwuid(os.getuid())
shell_path = os.getenv("SHELL", struct_passwd.pw_shell)
shell = shell_path.rsplit("/", 1)[-1]
user = struct_passwd.pw_name

# the *original* user's environment, readonly
user_environ: typing.Mapping[str, str] = os.environ.copy()

home = user_environ["HOME"] if CI else struct_passwd.pw_dir
root = f"{home}/.local/share/sentry-devenv"

homebrew_repo = "/opt/homebrew"
homebrew_bin = f"{homebrew_repo}/bin"
if INTEL_MAC:
    homebrew_repo = "/usr/local/Homebrew"
    homebrew_bin = "/usr/local/bin"

# compatibility with devenv <= 1.4.0
# (used in sentry sync.py)
VOLTA_HOME = f"{root}/volta"



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TypeAlias

from devenv.lib.context import Context

ExitCode: TypeAlias = "str | int | None"

Action: TypeAlias = "Callable[[Context, Sequence[str] | None], ExitCode]"


@dataclass(frozen=True)
class DevModuleInfo:
    name: str
    command: str
    help: str
    action: Action


def require(var: str, message: str) -> Callable[[Action], Action]:
    def outer(main: Action) -> Action:
        def inner(context: Context, args: Sequence[str] | None) -> ExitCode:
            if context.get(var) is None:
                raise SystemExit(message)
            return main(context, args)

        return inner

    return outer


require_repo = require("repo", "This command requires a repository")



from __future__ import annotations

import os
import shutil
import tempfile
from typing import Optional

from devenv.constants import home
from devenv.constants import root
from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive.download(url, sha256, dest=f"{tmpd}/colima")
        os.replace(f"{tmpd}/colima", f"{into}/colima")
        os.chmod(f"{into}/colima", 0o775)


def uninstall(binroot: str) -> None:
    for d in (f"{home}/.lima",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (f"{binroot}/colima",):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def install(
    version: str, url: str, sha256: str, reporoot: Optional[str] = ""
) -> None:
    if reporoot:
        binroot = fs.ensure_binroot(reporoot)
    else:
        # compatibility with devenv <= 1.4.0
        binroot = f"{root}/bin"
        os.makedirs(binroot, exist_ok=True)

    if shutil.which("colima", path=binroot) == f"{binroot}/colima":
        stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
        installed_version = stdout.strip().split()[-1]
        if version == installed_version:
            return
        print(f"installed colima {installed_version} is outdated!")

    print(f"installing colima {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
    if f"colima version {version}" not in stdout:
        raise SystemExit(f"Failed to install colima {version}! Found: {stdout}")



from __future__ import annotations

import os

from devenv.constants import home
from devenv.constants import shell
from devenv.lib import proc


def shellrc() -> str:
    if shell == "zsh":
        return f"{home}/.zshrc"
    if shell == "bash":
        return f"{home}/.bashrc"
    if shell == "fish":
        return f"{home}/.config/fish/config.fish"
    raise NotImplementedError(f"unsupported shell: {shell}")


def gitroot(cd: str = "") -> str:
    from os.path import normpath, join

    if not cd:
        cd = os.getcwd()

    stdout = proc.run(
        ("git", "-C", cd, "rev-parse", "--show-cdup"), stdout=True
    )
    return normpath(join(cd, stdout))


def idempotent_add(filepath: str, text: str) -> None:
    if not os.path.exists(filepath):
        with open(filepath, "w") as f:
            f.write(f"\n{text}\n")
            return
    with open(filepath, "r+") as f:
        contents = f.read()
        if text not in contents:
            f.write(f"\n{text}\n")


def write_script(filepath: str, text: str) -> None:
    with open(filepath, "w") as f:
        f.write(text)
    os.chmod(filepath, 0o775)


def ensure_binroot(reporoot: str) -> str:
    binroot = f"{reporoot}/.devenv/bin"
    os.makedirs(binroot, exist_ok=True)
    if not os.path.exists(f"{binroot}/.gitignore"):
        with open(f"{binroot}/.gitignore", "w") as f:
            f.write(
                """*
# automatically written by devenv ensure_binroot! feel free to modify.
"""
            )
    return binroot


def ensure_symlink(expected_src: str, dest: str) -> None:
    try:
        src = os.readlink(dest)
        if src != expected_src:
            print(f"WARNING: {dest} unexpectedly points to {src}")
            return
    except FileNotFoundError:
        os.symlink(expected_src, dest)
    except OSError as e:
        if e.errno == 22:
            print(f"WARNING: {dest} exists and isn't a symlink")
            return



from __future__ import annotations

from unittest.mock import call
from unittest.mock import patch

from devenv.constants import home
from devenv.constants import root
from devenv.lib.direnv import _sha256
from devenv.lib.direnv import _version
from devenv.lib.direnv import install


# imo this is enough coverage (as in just keeping it to bash + darwin-arm64)
# we just want to verify that the important functions were called like fs_idempotent_add
def test_install() -> None:
    with patch("devenv.lib.direnv.MACHINE", "arm64"), patch(
        "devenv.lib.direnv.sys.platform", "darwin"
    ), patch("devenv.lib.archive.download") as mock_archive_download, patch(
        "devenv.lib.fs.idempotent_add"
    ) as mock_lib_fs_idempotent_add, patch(
        "os.chmod"
    ), patch(
        "shutil.which", return_value=False
    ), patch(
        "devenv.lib.direnv.proc.run", side_effect=[_version]  # direnv version
    ) as mock_proc_run:
        install()
        assert mock_lib_fs_idempotent_add.mock_calls == [
            # SHELL is set in conftest.py
            call(f"{home}/.bashrc", '\neval "$(direnv hook bash)"\n')
        ]
        assert mock_archive_download.mock_calls == [
            call(
                f"https://github.com/direnv/direnv/releases/download/v{_version}/direnv.darwin-arm64",
                _sha256["direnv.darwin-arm64"],
                dest=f"{root}/bin/direnv",
            )
        ]
        assert mock_proc_run.mock_calls == [
            call((f"{root}/bin/direnv", "version"), stdout=True)
        ]



from __future__ import annotations

import argparse
import re
import subprocess
from collections.abc import Sequence
from functools import lru_cache

from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo


@lru_cache(maxsize=None)
def get_sha(repo: str, ref: str) -> str:
    if len(ref) == 40:
        try:
            int(ref, 16)
        except ValueError:
            pass
        else:
            return ref

    cmd = ("git", "ls-remote", "--exit-code", f"https://github.com/{repo}", ref)
    out = subprocess.check_output(cmd)
    for line in out.decode().splitlines():
        sha, refname = line.split()
        if refname in (f"refs/tags/{ref}", f"refs/heads/{ref}"):
            return sha
    else:
        raise AssertionError(f"unknown ref: {repo}@{ref}")


def extract_repo(action: str) -> str:
    # Some actions can be like `github/codeql-action/init`,
    # where init is just a directory. The ref is for the whole repo.
    # We only want the repo name though.
    parts = action.split("/")
    return f"{parts[0]}/{parts[1]}"


def main(context: Context, argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "files", nargs="+", type=str, help="path to github actions file"
    )
    args = parser.parse_args(argv)
    files: list[str] = args.files

    ACTION_VERSION_RE = re.compile(
        r"(?<=uses: )(?P<action>.*)@(?P<ref>[^#\s]+)"
    )

    for fp in files:
        with open(fp, "r+") as f:
            newlines: list[str] = []
            for line in f:
                m = ACTION_VERSION_RE.search(line)
                if not m:
                    newlines.append(line)
                    continue
                d = m.groupdict()
                sha = get_sha(extract_repo(d["action"]), ref=d["ref"])
                if sha != d["ref"]:
                    line = ACTION_VERSION_RE.sub(rf"\1@{sha} # \2", line)
                newlines.append(line)
            f.seek(0)
            f.truncate()
            f.writelines(newlines)

    return 0


module_info = DevModuleInfo(
    action=main, name=__name__, command="pin_gha", help="Pins github actions."
)



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

from devenv import doctor
from tests.doctor.devenv.checks import broken_fix
from tests.doctor.devenv.checks import failing_check
from tests.doctor.devenv.checks import failing_check_with_msg
from tests.doctor.devenv.checks import passing_check


def test_attempt_fix_success() -> None:
    check = doctor.Check(passing_check)
    assert doctor.attempt_fix(check, ThreadPoolExecutor()) == (True, "")


def test_attempt_fix_failure() -> None:
    check = doctor.Check(failing_check)
    assert doctor.attempt_fix(check, ThreadPoolExecutor()) == (False, "")


def test_attempt_fix_failure_with_msg() -> None:
    check = doctor.Check(failing_check_with_msg)
    assert doctor.attempt_fix(check, ThreadPoolExecutor()) == (
        False,
        "fix failed",
    )


def test_attempt_fix_broken_fix() -> None:
    check = doctor.Check(broken_fix)
    assert doctor.attempt_fix(check, ThreadPoolExecutor()) == (
        False,
        "Fix threw a runtime exception: division by zero",
    )



from __future__ import annotations

import hashlib
import os
import secrets
import shutil
import tarfile
import tempfile
import urllib.request
from urllib.error import HTTPError

from devenv.constants import home


def atomic_replace(src: str, dest: str) -> None:
    if os.path.dirname(src) != os.path.dirname(dest):
        raise RuntimeError(
            f"cannot atomically move to dest {dest}; it needs to be in the same dir as {src}"
        )
    os.replace(src, dest)


def download(url: str, sha256: str, dest: str = "") -> str:
    if not dest:
        cache_root = f"{home}/.cache/sentry-devenv"
        dest = f"{cache_root}/{sha256}"
        os.makedirs(cache_root, exist_ok=True)

    if not os.path.exists(dest):
        try:
            resp = urllib.request.urlopen(url)
        except HTTPError as e:
            raise RuntimeError(f"Error getting {url}: {e}")

        dest_dir = os.path.dirname(dest)
        os.makedirs(dest_dir, exist_ok=True)

        with tempfile.NamedTemporaryFile(delete=False, dir=dest_dir) as tmpf:
            shutil.copyfileobj(resp, tmpf)
            tmpf.seek(0)
            checksum = hashlib.sha256()
            buf = tmpf.read(4096)
            while buf:
                checksum.update(buf)
                buf = tmpf.read(4096)

            if not secrets.compare_digest(checksum.hexdigest(), sha256):
                raise RuntimeError(
                    f"checksum mismatch for {url}:\n"
                    f"- got: {checksum.hexdigest()}\n"
                    f"- expected: {sha256}\n"
                )

            atomic_replace(tmpf.name, dest)

    return dest


def unpack(path: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tarfile.open(name=path, mode="r:*") as tarf:
        tarf.extractall(into)



from __future__ import annotations

from devenv import doctor
from tests.doctor.devenv.checks import failing_check
from tests.doctor.devenv.checks import passing_check


def test_filter_failing_checks_no_checks() -> None:
    assert doctor.filter_failing_checks({}) == []


def test_filter_failing_checks_one_passing_check() -> None:
    check = doctor.Check(passing_check)
    assert doctor.filter_failing_checks({check: (True, "")}) == []


def test_filter_failing_checks_one_failing_check() -> None:
    check = doctor.Check(failing_check)
    assert doctor.filter_failing_checks({check: (False, "")}) == [check]


def test_filter_failing_checks_one_passing_and_one_failing_check() -> None:
    first_check = doctor.Check(passing_check)
    second_check = doctor.Check(failing_check)
    assert doctor.filter_failing_checks(
        {first_check: (True, ""), second_check: (False, "")}
    ) == [second_check]


def test_filter_failing_checks_no_duplicate_checks() -> None:
    first_check = doctor.Check(passing_check)
    second_check = doctor.Check(failing_check)
    assert doctor.filter_failing_checks(
        {
            first_check: (True, ""),
            second_check: (False, ""),
            second_check: (False, ""),
        }
    ) == [second_check]



from __future__ import annotations

import sys
from unittest.mock import patch

from devenv import pythons
from devenv.constants import root


def test_get() -> None:
    with (
        patch("devenv.lib.archive.download"),
        patch("devenv.lib.archive.unpack"),
        patch("os.path.exists", side_effect=[False, True]),
    ):
        path = pythons.get("3.12.0", "foo", "bar")
        assert path == f"{root}/pythons/3.12.0/python/bin/python3"


def test_get_system() -> None:
    v = sys.version_info
    python_version = f"{v.major}.{v.minor}.{max(v.micro - 1, 0)}"

    with patch("devenv.pythons.FORCE_PY", False):
        path = pythons.get(python_version, "foo", "bar")
        assert path == sys.executable



from __future__ import annotations

from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = set(["test", "fail"])
name = "failing check with msg"


@checker
def check() -> tuple[bool, str]:
    return False, "check failed"


@fixer
def fix() -> tuple[bool, str]:
    return False, "fix failed"



from __future__ import annotations

import os
import pathlib
from unittest.mock import call
from unittest.mock import patch

from devenv.lib import config
from devenv.lib import venv
from devenv.lib.repository import Repository

mock_config = """
[venv.sentry-kube]
python = 3.11.6
requirements = k8s/cli/requirements.txt
editable =
  k8s/cli
  k8s/cli/libsentrykube
bins =
  pre-commit
  pyupgrade
  sentry-kube
  sentry-kube-pop

[python3.11.6]
darwin_x86_64 = https://github.com/indygreg/python-build-standalone/releases/download/20231002/cpython-3.11.6+20231002-x86_64-apple-darwin-install_only.tar.gz
darwin_x86_64_sha256 = 178cb1716c2abc25cb56ae915096c1a083e60abeba57af001996e8bc6ce1a371
darwin_arm64 = https://github.com/indygreg/python-build-standalone/releases/download/20231002/cpython-3.11.6+20231002-aarch64-apple-darwin-install_only.tar.gz
darwin_arm64_sha256 = 916c35125b5d8323a21526d7a9154ca626453f63d0878e95b9f613a95006c990
linux_x86_64 = https://github.com/indygreg/python-build-standalone/releases/download/20231002/cpython-3.11.6+20231002-x86_64-unknown-linux-gnu-install_only.tar.gz
linux_x86_64_sha256 = ee37a7eae6e80148c7e3abc56e48a397c1664f044920463ad0df0fc706eacea8
linux_arm64 = https://github.com/indygreg/python-build-standalone/releases/download/20231002/cpython-3.11.6+20231002-aarch64-unknown-linux-gnu-install_only.tar.gz
linux_arm64_sha256 = 3e26a672df17708c4dc928475a5974c3fb3a34a9b45c65fb4bd1e50504cc84ec
"""


def test_get_ensure(tmp_path: pathlib.Path) -> None:
    repo = Repository(f"{tmp_path}/ops")

    os.makedirs(repo.config_path)
    with open(f"{repo.config_path}/config.ini", "w") as f:
        f.write(mock_config)

    venv_dir, python_version, requirements, editable_paths, bins = venv.get(
        repo.path, "sentry-kube"
    )
    os.makedirs(venv_dir)

    assert (venv_dir, python_version, requirements, editable_paths, bins) == (
        f"{repo.path}/.venv-sentry-kube",
        "3.11.6",
        f"{repo.path}/k8s/cli/requirements.txt",
        (f"{repo.path}/k8s/cli", f"{repo.path}/k8s/cli/libsentrykube"),
        ("pre-commit", "pyupgrade", "sentry-kube", "sentry-kube-pop"),
    )

    url, sha256 = config.get_python(repo.path, python_version)

    with patch("devenv.lib.venv.proc.run") as mock_run, patch(
        "devenv.lib.venv.pythons.get", return_value="python"
    ) as mock_pythons_get, patch("shutil.rmtree"):
        venv.ensure(venv_dir, python_version, url, sha256)
        assert mock_pythons_get.mock_calls == [
            call(python_version, url, sha256)
        ]
        assert mock_run.mock_calls == [
            call(("python", "-m", "venv", venv_dir), exit=True)
        ]
        assert os.path.isfile(f"{venv_dir}/.gitignore")

    # fake venv
    with open(f"{venv_dir}/pyvenv.cfg", "w") as f:
        f.write(f"version = {python_version}\n")

    assert venv.check(venv_dir, python_version) == venv.VenvStatus.OK

    # unittesting venv.sync(venv_dir, requirements, editable_paths, bins)
    # isn't really useful and is covered better with integration



from __future__ import annotations

import os
import platform
import tempfile
from shutil import which
from typing import Optional

from devenv.constants import MACHINE
from devenv.constants import root
from devenv.lib import archive
from devenv.lib import fs

_version = "0.19.1"
_sha256 = {
    f"lima-{_version}-Darwin-arm64.tar.gz": "0dfcf3a39782baf1c2ea43cf026f8df0321c671d914c105fbb78de507aa8bda4",
    f"lima-{_version}-Darwin-x86_64.tar.gz": "ac8827479f66ef1b288b31f164b22f6433faa14c44ce5bbebe09e6e913582479",
    f"lima-{_version}-Linux-aarch64.tar.gz": "c55e57ddbefd9988d0f3676bb873bcc6e0f7b3c3d47a1f07599ee151c5198d96",
    f"lima-{_version}-Linux-x86_64.tar.gz": "7d18b1716aae14bf98d6ea93a703e8877b0c3142f7ba2e87401d47d5d0fe3ff1",
}


class UnexpectedPlatformError(Exception):
    pass


def build_name() -> str | None:
    system = platform.system()
    if system == "Linux":
        if MACHINE == "arm64":
            return f"lima-{_version}-{system}-aarch64.tar.gz"
        return f"lima-{_version}-{system}-{MACHINE}.tar.gz"
    elif system == "Darwin":
        return f"lima-{_version}-{system}-{MACHINE}.tar.gz"
    else:
        raise UnexpectedPlatformError(f"Unexpected OS: {platform.platform()}")


def download_and_unpack_archive(name: str, into: str) -> None:
    url = (
        "https://github.com/lima-vm/lima/releases/download/"
        f"v{_version}/{name}"
    )

    archive_file = archive.download(url, _sha256[name], dest=f"{into}/{name}")
    archive.unpack(archive_file, into)


def _install(into: str) -> None:
    name = build_name()
    if name is None:
        return

    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        download_and_unpack_archive(name, tmpd)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/bin/lima", f"{into}/lima")
        os.replace(f"{tmpd}/bin/limactl", f"{into}/limactl")
        os.replace(
            f"{tmpd}/share/lima/lima-guestagent.Linux-aarch64",
            f"{into}/lima-guestagent.Linux-aarch64",
        )
        os.replace(
            f"{tmpd}/share/lima/lima-guestagent.Linux-x86_64",
            f"{into}/lima-guestagent.Linux-x86_64",
        )
        os.makedirs(f"{into}/templates", exist_ok=True)
        os.replace(
            f"{tmpd}/share/lima/templates/default.yaml",
            f"{into}/templates/default.yaml",
        )


def install(reporoot: Optional[str] = "") -> None:
    if reporoot:
        binroot = fs.ensure_binroot(reporoot)
    else:
        # compatibility with devenv <= 1.4.0
        binroot = f"{root}/bin"
        os.makedirs(binroot, exist_ok=True)

    # this needs to be better
    if (
        which("lima", path=binroot) == f"{binroot}/lima"
        and which("limactl", path=binroot) == f"{binroot}/limactl"
    ):
        return

    # TODO: uninstall and repolocal config for versions
    _install(binroot)

    if not os.path.exists(f"{binroot}/lima") or not os.path.exists(
        f"{binroot}/limactl"
    ):
        raise SystemExit("Failed to install colima!")



from __future__ import annotations

import argparse
import os
from collections.abc import Sequence

from devenv import bootstrap
from devenv import doctor
from devenv import fetch
from devenv import pin_gha
from devenv import sync
from devenv.constants import home
from devenv.constants import user
from devenv.constants import version
from devenv.lib.config import read_config
from devenv.lib.context import Context
from devenv.lib.fs import gitroot
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import ExitCode
from devenv.lib.repository import Repository


def devenv(argv: Sequence[str], config_path: str) -> ExitCode:
    # determine current repo, if applicable
    fake_reporoot = os.getenv("CI_DEVENV_INTEGRATION_FAKE_REPOROOT")
    if fake_reporoot:
        current_root = fake_reporoot
    else:
        try:
            current_root = gitroot()
        except RuntimeError:
            current_root = None

    # This may or may not exist
    config = read_config(config_path)

    # Guessing temporary code root
    code_root = config.get("devenv", "coderoot", fallback=None) or (
        os.path.abspath(f"{current_root}/..")
        if current_root
        else os.path.expanduser("~/code")
    )

    modinfo_list: Sequence[DevModuleInfo] = [
        module.module_info
        for module in [bootstrap, fetch, doctor, pin_gha, sync]
        if hasattr(module, "module_info")
    ]

    # TODO: Search for modules in work repo

    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=version)
    subparser = parser.add_subparsers(
        title=argparse.SUPPRESS,
        metavar="command",
        dest="command",
        required=True,
    )

    for info in modinfo_list:
        # Argparse stuff
        subparser.add_parser(info.command, help=info.help)

    if len(argv) == 1:
        parser.print_help()
        return 0

    args, remainder = parser.parse_known_args(argv[1:])

    # context for subcommands
    context: Context = {
        "config_path": config_path,
        "code_root": code_root,
        "repo": Repository(current_root) if current_root else None,
    }

    command_actions = {info.command: info.action for info in modinfo_list}
    action = command_actions.get(args.command)
    assert action is not None
    return action(context, remainder)


def main() -> ExitCode:
    if user == "root":
        raise SystemExit("You shouldn't be running devenv as root.")

    import sys
    import sentry_sdk

    sentry_sdk.init(
        # https://sentry.sentry.io/settings/projects/sentry-dev-env/keys/
        dsn="https://9bdb053cb8274ea69231834d1edeec4c@o1.ingest.sentry.io/5723503",
        # enable performance monitoring
        enable_tracing=True,
    )

    return devenv(sys.argv, f"{home}/.config/sentry-devenv/config.ini")


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import os
from unittest.mock import call
from unittest.mock import patch

import pytest

from devenv.lib.volta import _install
from devenv.lib.volta import _sha256
from devenv.lib.volta import _version
from devenv.lib.volta import install
from devenv.lib.volta import populate_volta_home_with_shims
from devenv.lib.volta import UnexpectedPlatformError


def test_install(tmp_path: str) -> None:
    reporoot = f"{tmp_path}/reporoot"
    binroot = f"{reporoot}/.devenv/bin"
    VOLTA_HOME = f"{binroot}/volta-home"

    os.makedirs(f"{VOLTA_HOME}/bin")
    for executable in ("node", "npm", "npx", "yarn", "pnpm"):
        open(f"{VOLTA_HOME}/bin/{executable}", "a").close()

    with patch("shutil.which", side_effect=[None, None]), patch(
        "devenv.lib.volta.uninstall"  # need this so we don't undo the previous mock install
    ), patch("devenv.lib.volta._install") as mock_install_volta, patch(
        "devenv.lib.volta.proc.run",
        side_effect=[None, _version],  # volta-migrate  # volta -v
    ) as mock_proc_run:
        install(reporoot)

        mock_install_volta.assert_called_once_with(binroot)
        assert mock_proc_run.mock_calls == [
            call((f"{binroot}/volta-migrate",), env={"VOLTA_HOME": VOLTA_HOME}),
            call(
                (f"{binroot}/volta", "-v"),
                env={"VOLTA_HOME": VOLTA_HOME},
                stdout=True,
            ),
        ]

    # now test already installed
    with patch("devenv.lib.volta._install") as mock_install_volta, patch(
        "shutil.which",
        side_effect=[f"{binroot}/volta", f"{VOLTA_HOME}/bin/node"],
    ):
        install(reporoot)
        assert not mock_install_volta.called


def test_install_volta_linux_x86_64() -> None:
    with patch("platform.system", return_value="Linux") as mock_system, patch(
        "platform.machine", return_value="x86_64"
    ) as mock_machine, patch(
        "devenv.lib.volta.archive.download"
    ) as mock_download, patch(
        "devenv.lib.volta.archive.unpack"
    ) as mock_unpack:
        _install("/path/to/unpack")
        mock_system.assert_called_once()
        mock_machine.assert_called_once()
        mock_download.assert_called_once_with(
            (
                "https://github.com/volta-cli/volta/releases/download/"
                f"v{_version}/volta-{_version}-linux.tar.gz"
            ),
            _sha256[f"volta-{_version}-linux.tar.gz"],
        )
        mock_unpack.assert_called_once_with(
            mock_download.return_value, "/path/to/unpack"
        )


def test_install_volta_macos_x86_64() -> None:
    with patch("platform.system", return_value="Darwin") as mock_system, patch(
        "platform.machine", return_value="x86_64"
    ) as mock_machine, patch(
        "devenv.lib.volta.archive.download"
    ) as mock_download, patch(
        "devenv.lib.volta.archive.unpack"
    ) as mock_unpack:
        _install("/path/to/unpack")
        mock_system.assert_called_once()
        mock_machine.assert_called_once()
        mock_download.assert_called_once_with(
            (
                "https://github.com/volta-cli/volta/releases/download/"
                f"v{_version}/volta-{_version}-macos.tar.gz"
            ),
            _sha256[f"volta-{_version}-macos.tar.gz"],
        )
        mock_unpack.assert_called_once_with(
            mock_download.return_value, "/path/to/unpack"
        )


def test_install_volta_macos_arm64() -> None:
    with patch("platform.system", return_value="Darwin") as mock_system, patch(
        "platform.machine", return_value="arm64"
    ) as mock_machine, patch(
        "devenv.lib.volta.archive.download"
    ) as mock_download, patch(
        "devenv.lib.volta.archive.unpack"
    ) as mock_unpack:
        _install("/path/to/unpack")
        mock_system.assert_called_once()
        mock_machine.assert_called_once()
        mock_download.assert_called_once_with(
            (
                "https://github.com/volta-cli/volta/releases/download/"
                f"v{_version}/volta-{_version}-macos-aarch64.tar.gz"
            ),
            _sha256[f"volta-{_version}-macos-aarch64.tar.gz"],
        )
        mock_unpack.assert_called_once_with(
            mock_download.return_value, "/path/to/unpack"
        )


def test_install_volta_unexpected_platform() -> None:
    with patch("platform.system", return_value="Unexpected"):
        with pytest.raises(UnexpectedPlatformError):
            _install("/path/to/unpack")


def test_populate_volta_home_with_shims() -> None:
    with patch("devenv.lib.volta.proc.run") as mock_run:
        unpack_into = "/path/to/unpack"
        mock_run.side_effect = [
            None,  # mock run for volta-migrate
            _version,  # mock run for volta -v
        ]

        populate_volta_home_with_shims(unpack_into, "foo")

        assert mock_run.mock_calls == [
            call((f"{unpack_into}/volta-migrate",), env={"VOLTA_HOME": "foo"}),
            call(
                (f"{unpack_into}/volta", "-v"),
                env={"VOLTA_HOME": "foo"},
                stdout=True,
            ),
        ]



from __future__ import annotations

import configparser
import functools
import os
import sys
from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeAlias

from devenv.constants import CI
from devenv.constants import MACHINE


@dataclass(frozen=True)
class ConfigOpt:
    name: str
    prompt: str
    formatter: Callable[[str], str] | None = None
    default: Callable[[], str] = lambda: ""


def _path_formatter(path: str) -> str:
    return os.path.normpath(os.path.expanduser(path))


CONFIG_OPTS = {
    "coderoot": ConfigOpt(
        "coderoot",
        "# please enter the root directory you want to work in",
        _path_formatter,
        lambda: "~/code",
    )
}

Config: TypeAlias = "dict[str, dict[str, str | None]]"


def initialize_config(config_path: str, defaults: Config) -> None:
    config = configparser.ConfigParser()

    # Read existing configuration, if present
    if os.path.exists(config_path):
        config.read(config_path)
    else:
        os.makedirs(os.path.dirname(config_path), exist_ok=True)

    for section, values in defaults.items():
        if not config.has_section(section):
            config.add_section(section)

        for var, _val in values.items():
            opts = CONFIG_OPTS.get(var)

            val: str = config.get(
                section,
                var,
                fallback=(opts.default() if opts else "")
                if _val is None
                else _val,
            )

            if not CI:
                try:
                    if opts:
                        print(opts.prompt)
                    else:
                        print(f"{var}?")

                    val = input(f" [{val}]: ") or val
                except EOFError:
                    # noninterative, use the defaults
                    print()
            config.set(
                section,
                var,
                opts.formatter(val) if opts and opts.formatter else val,
            )

    print("Thank you. Saving answers.")

    with open(config_path, "w") as f:
        config.write(f)
    print(f"If you made a mistake, you can edit {config_path}.")


def read_config(path: str) -> configparser.ConfigParser:
    """Reads a configuration file from disk, with no caching"""
    config = configparser.ConfigParser()
    config.read(path)
    return config


@functools.lru_cache(maxsize=None)
def get_config(path: str) -> configparser.ConfigParser:
    """Reads a configuration file from disk, with caching"""
    config = configparser.ConfigParser()
    config.read(path)
    return config


def get_repo(reporoot: str) -> configparser.ConfigParser:
    """Deprecated"""
    from devenv.lib.repository import Repository

    return Repository(reporoot).config()


def get_python(reporoot: str, python_version: str) -> tuple[str, str]:
    cfg = get_repo(reporoot)
    url = cfg[f"python{python_version}"][f"{sys.platform}_{MACHINE}"]
    sha256 = cfg[f"python{python_version}"][f"{sys.platform}_{MACHINE}_sha256"]
    return url, sha256


# only used for sentry/getsentry
def get_python_legacy(reporoot: str, python_version: str) -> tuple[str, str]:
    cfg = get_repo(reporoot)
    url = cfg["python"][f"{sys.platform}_{MACHINE}"]
    sha256 = cfg["python"][f"{sys.platform}_{MACHINE}_sha256"]
    return url, sha256



from __future__ import annotations

import argparse
import os
import shutil
from collections.abc import Sequence

from devenv.constants import CI
from devenv.constants import EXTERNAL_CONTRIBUTOR
from devenv.lib import brew
from devenv.lib import direnv
from devenv.lib import github
from devenv.lib import proc
from devenv.lib.config import Config
from devenv.lib.config import initialize_config
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import ExitCode


def main(context: Context, argv: Sequence[str] | None = None) -> ExitCode:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--default-config",
        action="append",
        help="Provide a default config value. e.g., -d coderoot:path/to/root",
    )

    args = parser.parse_args(argv)

    configs = {
        k: v for k, v in [i.split(":", 1) for i in args.default_config or []]
    }

    if "coderoot" not in configs and "code_root" in context:
        configs["coderoot"] = context["code_root"]

    default_config: Config = {"devenv": configs}
    initialize_config(context["config_path"], default_config)

    if not CI and shutil.which("xcrun"):
        # xcode-select --install will take a while,
        # and involves elevated permissions with a GUI,
        # so best to just let the user go through that separately then retrying,
        # rather than waiting for it.
        # There is a way to perform a headless install but it's more complex
        # (refer to how homebrew does it).
        try:
            _ = proc.run(("xcrun", "-f", "git"), stdout=True)
        except RuntimeError:
            return "Failed to find git. Run xcode-select --install, then re-run bootstrap when done."

    github.add_to_known_hosts()

    if not EXTERNAL_CONTRIBUTOR and not github.check_ssh_access():
        is_employee = (
            False
            if CI
            else input("Are you a Sentry employee? (Y/n): ").lower()
            in {"y", "yes", ""}
        )
        if not CI and not is_employee:
            print(
                "Please set the SENTRY_EXTERNAL_CONTRIBUTOR environment variable and re-run bootstrap."
            )
            return 1
        pubkey = github.generate_and_configure_ssh_keypair()
        if not CI:
            input(
                f"""
Failed to authenticate with an ssh key to GitHub.
We've generated and configured one for you at ~/.ssh/sentry-github.
Visit https://github.com/settings/ssh/new and add the following Authentication key:

{pubkey}

Then, you need to go to https://github.com/settings/keys, find your key,
and click Configure SSO, for the getsentry organization.

When done, hit ENTER to continue.
"""
            )
        while not github.check_ssh_access():
            input(
                "Still failing to authenticate to GitHub. ENTER to retry, otherwise ^C to quit."
            )

    brew.install()
    direnv.install()

    os.makedirs(context["code_root"], exist_ok=True)

    print(
        """
All done! Please close this terminal window and start a fresh one.

Afterward, start working on your project using the devenv fetch command
e.g., devenv fetch sentry or devenv fetch ops
"""
    )

    return 0


module_info = DevModuleInfo(
    action=main,
    name=__name__,
    command="bootstrap",
    help="Bootstraps the development environment.",
)



from __future__ import annotations

import os

from devenv.lib import proc


def packages() -> list[str]:
    # note: brew leaves will not print out top-level casks unfortunately
    stdout = proc.run(
        ("brew", "list"),
        exit=True,
        env={
            **os.environ,
            "HOMEBREW_NO_AUTO_UPDATE": "1",
            "HOMEBREW_NO_INSTALLED_DEPENDENTS_CHECK": "1",
            "HOMEBREW_NO_INSTALL_CLEANUP": "1",
            "HOMEBREW_NO_ANALYTICS": "1",
        },
        stdout=True,
    )
    return stdout.split()



from __future__ import annotations

from unittest.mock import patch

from devenv.lib.brew import add_brew_to_shellrc
from devenv.lib.brew import create_dirs
from devenv.lib.brew import install


def test_install_brew_already_installed() -> None:
    with patch(
        "devenv.lib.brew.which", return_value="/usr/local/bin/brew"
    ), patch(
        "devenv.lib.brew.add_brew_to_shellrc"
    ) as mock_add_brew_to_shellrc, patch(
        "devenv.lib.brew.proc.run"
    ) as mock_run, patch(
        "os.symlink"
    ) as mock_symlink:
        install()
        mock_run.assert_not_called()
        mock_run.assert_not_called()
        mock_symlink.assert_not_called()
        mock_add_brew_to_shellrc.assert_not_called()


def test_install_brew_not_installed_intel_mac() -> None:
    # Mock the necessary dependencies
    homebrew_repo = "/path/to/homebrew_repo"
    homebrew_bin = "/path/to/homebrew/bin"

    with patch("devenv.lib.brew.which", return_value=None), patch(
        "devenv.lib.brew.create_dirs"
    ) as mock_create_dirs, patch(
        "devenv.lib.brew.add_brew_to_shellrc"
    ) as mock_add_brew_to_shellrc, patch(
        "devenv.lib.brew.INTEL_MAC", True
    ), patch(
        "devenv.lib.brew.DARWIN", True
    ), patch(
        "devenv.lib.brew.proc.run"
    ) as mock_run, patch(
        "os.symlink"
    ) as mock_symlink, patch.multiple(
        "devenv.lib.brew",
        homebrew_repo=homebrew_repo,
        homebrew_bin=homebrew_bin,
    ):
        install()
        mock_create_dirs.assert_called_once()
        mock_run.assert_called_once_with(
            (
                "git",
                "-C",
                homebrew_repo,
                "clone",
                "--depth=1",
                "https://github.com/Homebrew/brew",
                ".",
            )
        )
        mock_symlink.assert_called_once_with(
            f"{homebrew_repo}/bin/brew", f"{homebrew_bin}/brew"
        )
        mock_add_brew_to_shellrc.assert_called_once()


def test_install_brew_not_installed_non_intel_mac() -> None:
    # Mock the necessary dependencies
    homebrew_repo = "/path/to/homebrew_repo"
    homebrew_bin = "/path/to/homebrew/bin"

    with patch("devenv.lib.brew.which", return_value=None), patch(
        "devenv.lib.brew.create_dirs"
    ) as mock_create_dirs, patch(
        "devenv.lib.brew.add_brew_to_shellrc"
    ) as mock_add_brew_to_shellrc, patch(
        "devenv.lib.brew.INTEL_MAC", False
    ), patch(
        "devenv.lib.brew.DARWIN", True
    ), patch(
        "devenv.lib.brew.proc.run"
    ) as mock_run, patch(
        "os.symlink"
    ) as mock_symlink, patch.multiple(
        "devenv.lib.brew",
        homebrew_repo=homebrew_repo,
        homebrew_bin=homebrew_bin,
    ):
        install()
        mock_create_dirs.assert_called_once()
        mock_run.assert_called_once_with(
            (
                "git",
                "-C",
                homebrew_repo,
                "clone",
                "--depth=1",
                "https://github.com/Homebrew/brew",
                ".",
            )
        )
        # We don't symlink brew on non-Intel Macs
        mock_symlink.assert_not_called()
        mock_add_brew_to_shellrc.assert_called_once()


def test_add_brew_to_shellrc() -> None:
    # Mock the necessary dependencies
    shellrc_path = "/path/to/shellrc"
    homebrew_bin = "/path/to/homebrew/bin"

    with patch(
        "devenv.lib.brew.fs.shellrc", return_value=shellrc_path
    ) as mock_shellrc, patch(
        "devenv.lib.brew.fs.idempotent_add"
    ) as mock_idempotent_add, patch(
        "devenv.lib.brew.homebrew_bin", homebrew_bin
    ):
        add_brew_to_shellrc()

        mock_shellrc.assert_called_once()
        mock_idempotent_add.assert_called_once_with(
            shellrc_path, f'\neval "$({homebrew_bin}/brew shellenv)"\n'
        )


def test_create_dirs() -> None:
    # Mock the necessary dependencies
    homebrew_repo = "/path/to/homebrew/repo"
    user = "test_user"
    INTEL_MAC = False

    # Mock the proc.run function
    with patch("devenv.lib.brew.proc.run") as mock_run, patch.multiple(
        "devenv.lib.brew",
        user=user,
        homebrew_repo=homebrew_repo,
        INTEL_MAC=INTEL_MAC,
    ):
        create_dirs()

        # Assert that proc.run was called with the correct arguments
        mock_run.assert_called_once_with(
            (
                "sudo",
                "bash",
                "-c",
                f"""
mkdir -p {homebrew_repo}
chown {user} {homebrew_repo}
""",
            ),
            exit=False,
        )


def test_create_dirs_intel_mac() -> None:
    # Mock the necessary dependencies
    homebrew_repo = "/path/to/homebrew/repo"
    user = "test_user"
    INTEL_MAC = True

    # Mock the proc.run function
    with patch("devenv.lib.brew.proc.run") as mock_run, patch.multiple(
        "devenv.lib.brew",
        user=user,
        homebrew_repo=homebrew_repo,
        INTEL_MAC=INTEL_MAC,
    ):
        create_dirs()

        # Assert that proc.run was called with the correct arguments
        mock_run.assert_called_once_with(
            (
                "sudo",
                "bash",
                "-c",
                f"""
mkdir -p {homebrew_repo} /usr/local/Cellar /usr/local/Caskroom /usr/local/Frameworks /usr/local/bin /usr/local/etc /usr/local/include /usr/local/lib /usr/local/opt /usr/local/sbin /usr/local/share /usr/local/var
chown {user} {homebrew_repo} /usr/local/Cellar /usr/local/Caskroom /usr/local/Frameworks /usr/local/bin /usr/local/etc /usr/local/include /usr/local/lib /usr/local/opt /usr/local/sbin /usr/local/share /usr/local/var
""",
            ),
            exit=False,
        )
